#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <string.h>
#include <unistd.h>
#include <wait.h>

#include <security/pam_appl.h>
#include <security/pam_modules.h>

#include "pam_ru.h"

int use_debug = 0, unbecome_mode = 0;

/*legacy become */
struct ruauthdata {
  char *ph_path;
  char *ru_agent;
};

int become_check(char *user, char *becomeacct, int ldap, int debug) {
  
    /*if (ldap) {
        if (debug) debout("ru_become:become_check: using ldap_check.");
        return ldap_check(becomeacct, user, debug);
    } 
    else {*/
        if (debug) debout("ru_become:become_check: using nss_check.");
        return nss_check(becomeacct, user, debug);
    //}
}

int is_become(char *becomeacct, int ldap, int debug) {

    /*if (ldap) {
        if (debug) debout("ru_become:is_become: using ldap_is_become.");
        return ldap_is_become(becomeacct, debug); 
    } 
    else {*/
        if (debug) debout("ru_become:is_become: using nss_is_become.");
        return nss_is_become(becomeacct, debug);
    //}
}

/*
* For those keeping score, user is the become account, and become user
* is the person trying to log in to the become account.  
*
* You may be asking... wouldn't it be better to just rename things?
*/

int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv) {
    
    int i, use_ldap = 0;
    char becomeuser[8192], *password, *user, *user_breakpt, *becomegroup;
    struct pam_conv *conv;
    struct pam_message *msgvec[1];
    struct pam_response *respvec[1];
    struct pam_message msg;
    
    char *group_h, *user_h;
    
    msgvec[0] = &msg; 
    
    for (i = 0; i < argc; i++) { 
    if (!strcmp("debug", argv[i])) {
            use_debug = 1;
        }
        else if (!strcmp("unbecome", argv[i])) {
            unbecome_mode = 1;
        }
    }
    
    if (use_debug) debout("become: getting conversation.");

#ifdef sun
    if (pam_get_item(pamh, PAM_CONV, (void **)&conv) != PAM_SUCCESS)
        return PAM_SERVICE_ERR;
#endif

#ifdef linux    
    if (pam_get_item(pamh, PAM_CONV, (const void **)&conv) != PAM_SUCCESS)
        return PAM_SERVICE_ERR;
#endif
    
    if (use_debug) debout("become: getting user.");

    if (!unbecome_mode) {

        if (use_debug) debout("become: become mode");
    
        if (pam_get_data(pamh, "become_group", (const void **)&becomegroup) == PAM_SUCCESS) {

            if (!becomegroup) {
                if (use_debug) debout("become: become_group already set, but it's NULL");
                return PAM_SERVICE_ERR;
            }            

            if (use_debug) debout("become: become_group already set, resetting PAM_USER");      

            if (pam_set_item(pamh, PAM_USER, becomegroup) != PAM_SUCCESS) {
                if (use_debug) debout("become: set_item PAM_USER failed");
                return PAM_SERVICE_ERR;
            } 
            else {
                user = strdup(becomegroup);
                if (use_debug) debout("become: set_item PAM_USER = %s", user);
            }

        }
        else {
            if (use_debug) debout("become: getting user.");

#ifdef sun
            if (pam_get_user(pamh, (char **)&user, NULL) != PAM_SUCCESS || !user || !*user) {
                return PAM_USER_UNKNOWN;
	    }
#endif

#ifdef linux
            if (pam_get_user(pamh, (const char **)&user, NULL) != PAM_SUCCESS || !user || !*user) {
                return PAM_USER_UNKNOWN;
            }
#endif

        }
    } 
    else if (pam_get_data(pamh, "become_group", (const void **)&becomegroup) == PAM_SUCCESS) {
    
        if (use_debug) debout("become: unbecome mode, user using become account");

        if (becomegroup) {
            if (pam_set_item(pamh, PAM_USER, becomegroup) != PAM_SUCCESS) {
                if (use_debug) debout("become: set_item PAM_USER failed");
                return PAM_SERVICE_ERR;
            } 
            if (use_debug) debout("become: set_item PAM_USER = %s", becomegroup);
            return PAM_SUCCESS;
        } 
        else { 
            if (use_debug) debout("become: retrieved become_group, but it's NULL");
            return PAM_SERVICE_ERR;
        }
    }
    else {
        if (use_debug) debout("become: unbecome mode, user not using become account");
        return PAM_SUCCESS;
    }
    
    if (use_debug) debout("become: getting password.");
    
    if (pam_get_item(pamh, PAM_AUTHTOK, (void *)&password) != PAM_SUCCESS || password == NULL) {
        
        msg.msg_style = PAM_PROMPT_ECHO_OFF;
        msg.msg = "Token: ";

#ifdef sun
        if ((*(conv->conv)) (1, (struct pam_message **)msgvec, respvec, conv->appdata_ptr) != PAM_SUCCESS) {
            return PAM_AUTH_ERR;
        }
#endif

#ifdef linux	
        if ((*(conv->conv)) (1, (const struct pam_message **)msgvec, respvec, conv->appdata_ptr) != PAM_SUCCESS) {
            return PAM_AUTH_ERR;
        }
#endif
    
        password = respvec[0]->resp;
    
        if (pam_set_item(pamh, PAM_AUTHTOK, (void *)password) != PAM_SUCCESS)                    
            return PAM_SERVICE_ERR;
    }
    
    if (strstr(password, ":") && is_become(user, use_ldap, use_debug)) {
        
        if (use_debug) debout("become: breaking up the password.");
    
        strncpy(becomeuser, password, sizeof(becomeuser)-1);
        becomeuser[sizeof(becomeuser)-1] = '\0';
        user_breakpt = strchr(becomeuser, ':');
        *user_breakpt = '\0';
        password = user_breakpt + 1;
    
        if (use_debug) debout("become: becomeuser = %s, password = %s", becomeuser, password);
    } 
    else {
        if (use_debug) debout("become: not a become account. moving on.");
        return PAM_SUCCESS;
    }
    
    if (use_debug) debout("become: calling become_check.");
    
    if (become_check(becomeuser, user, use_ldap, use_debug) == 0) {
        if (use_debug) debout("become: %s cannot access %s", becomeuser, user);
        return PAM_AUTH_ERR;
    } 
    else {
    
        group_h = strdup(user);
        if (pam_set_data(pamh, "become_group", group_h, NULL) != PAM_SUCCESS) {
            if (use_debug) debout("become: set_data become_group not SUCCESS");
            return PAM_SERVICE_ERR;
        }
    
        if (pam_set_item(pamh, PAM_USER, (void *)becomeuser) != PAM_SUCCESS) {
            if (use_debug) debout("become: set_item PAM_USER not SUCCESS");
            return PAM_SERVICE_ERR;
        }
    
        user_h = strdup(becomeuser);
        if (pam_set_data(pamh, "become_user", user_h, NULL) != PAM_SUCCESS) {
            if (use_debug) debout("become: set_data become_user not SUCCESS");
            return PAM_SERVICE_ERR;
        }
    
        if (pam_set_item(pamh, PAM_AUTHTOK, (void *)password) != PAM_SUCCESS){
            if (use_debug) debout("become: set_item PAM_AUTHTOK not SUCCESS");
            return PAM_SERVICE_ERR;
        }
    
        return PAM_SUCCESS;
    }
}

int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv) {
  
    char *acct;
    char *user;
  
    struct ruauthdata *ad;
  
    if (use_debug) debout("become: become setcred");
  
    if (pam_get_data(pamh, "become_group", (const void **)&acct) == PAM_SUCCESS) {
        
        if (use_debug) debout("become: become acct: %s", acct); 
    
        if (pam_set_item(pamh, PAM_USER, (void *)acct) != PAM_SUCCESS)
            return PAM_SERVICE_ERR;
    
        if (pam_get_data(pamh, "become_user", (const void **)&user) == PAM_SUCCESS) {
      
            ad = (struct ruauthdata *)malloc(sizeof (struct ruauthdata));
            if (ad) {
                ad->ph_path = NULL;
                ad->ru_agent = strdup(user);
                pam_set_data(pamh, "ru_auth", ad, NULL);
            } 
            else { 
                return PAM_SERVICE_ERR;
            }
        }
    } 
    else {
        if (use_debug) debout("become: normal account.  pass thru.");
    }
  
  return PAM_SUCCESS;
}
